from _typeshed import Incomplete
from typing import NamedTuple
from ujenkins.endpoints import Builds as Builds, Jobs as Jobs, Nodes as Nodes, System as System, Views as Views
from ujenkins.exceptions import JenkinsError as JenkinsError, JenkinsNotFoundError as JenkinsNotFoundError

class Response(NamedTuple):
    status: Incomplete
    headers: Incomplete
    body: Incomplete

class Jenkins:
    builds: Incomplete
    jobs: Incomplete
    nodes: Incomplete
    system: Incomplete
    views: Incomplete
    def __init__(self) -> None: ...



from ujenkins.endpoints.builds import Builds
from ujenkins.endpoints.jobs import Jobs
from ujenkins.endpoints.nodes import Nodes
from ujenkins.endpoints.system import System
from ujenkins.endpoints.views import Views

__all__ = (
    'Builds',
    'Jobs',
    'Nodes',
    'System',
    'Views',
)



import re

import responses

from tests import JENKINS_INFO_JSON

JENKINS_VIEW_CONFIG_XML = """
<?xml version="1.1" encoding="UTF-8"?>
<hudson.model.ListView>
  <name>my_view</name>
  <filterExecutors>false</filterExecutors>
  <filterQueue>false</filterQueue>
  <properties class="hudson.model.View$PropertyList"/>
  <jobNames>
    <comparator class="java.lang.String$CaseInsensitiveComparator"/>
  </jobNames>
  <jobFilters/>
  <columns>
    <hudson.views.StatusColumn/>
    <hudson.views.WeatherColumn/>
    <hudson.views.JobColumn/>
    <hudson.views.LastSuccessColumn/>
    <hudson.views.LastFailureColumn/>
    <hudson.views.LastDurationColumn/>
    <hudson.views.BuildButtonColumn/>
  </columns>
  <recurse>false</recurse>
</hudson.model.ListView>
"""


@responses.activate
def test_get(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/api/json'),
        body=JENKINS_INFO_JSON,
    )

    response = client.views.get()
    assert len(response) == 3
    assert 'my_view' in response


@responses.activate
def test_is_exists(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/api/json'),
        body=JENKINS_INFO_JSON,
    )

    assert client.views.is_exists('extra') is False


@responses.activate
def test_get_config(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/view/.+/config.xml'),
        content_type='application/xml',
        body=JENKINS_VIEW_CONFIG_XML,
    )

    config = client.views.get_config('buildbot')
    assert '<name>my_view</name>' in config


@responses.activate
def test_create(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/api/json'),
        body=JENKINS_INFO_JSON,
    )

    responses.add(
        responses.POST,
        re.compile(r'.*/createView'),
    )

    assert client.views.create('new_view', JENKINS_VIEW_CONFIG_XML) is None
    assert len(responses.calls) == 2


@responses.activate
def test_reconfigure(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/view/.+/config.xml'),
    )

    assert client.views.reconfigure('new_view', JENKINS_VIEW_CONFIG_XML) is None


@responses.activate
def test_delete(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/view/.+/doDelete'),
    )

    assert client.views.delete('new_view') is None



import json

from typing import Any, List, Optional, Union


class Builds:
    """
    List of Jenkins tags which can be used insted of build_id number.

    - lastBuild
    - lastCompletedBuild
    - lastFailedBuild
    - lastStableBuild
    - lastSuccessfulBuild
    - lastUnstableBuild
    - lastUnsuccessfulBuild
    """
    def __init__(self, jenkins):
        self.jenkins = jenkins

    def get(self, name: str) -> List[dict]:
        """
        Get list of builds for specified job.

        Example:

        .. code-block:: python

            builds = [
                {'number': 1, 'url': 'http://localhost/job/test/1/'},
                {'number': 2, 'url': 'http://localhost/job/test/2/'}
            ]

        Args:
            name (str):
                Job name or path (if in folder).

        Returns:
            List: list of build for specified job.
        """
        def callback(response) -> List[dict]:
            return json.loads(response.body)['allBuilds']

        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        return self.jenkins._request(
            'GET',
            f'/{folder_name}/job/{job_name}/api/json?tree=allBuilds[number,url]',
            _callback=callback,
        )

    def get_info(self, name: str, build_id: Union[int, str]) -> dict:
        """
        Get detailed information about specified build number of job.

        Args:
            name (str):
                Job name or path (if in folder).

            build_id (int):
                Build number or some of standard tags like `lastBuild`.

        Returns:
            dict: information about build.
        """
        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        return self.jenkins._request(
            'GET',
            f'/{folder_name}/job/{job_name}/{build_id}/api/json'
        )

    def get_output(self, name: str, build_id: Union[int, str]) -> str:
        """
        Get console output of specified build.

        Args:
            name (str):
                Job name or path (if in folder).

            build_id (int):
                Build number or some of standard tags like `lastBuild`.

        Returns:
            str: build output.
        """
        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        return self.jenkins._request(
            'GET',
            f'/{folder_name}/job/{job_name}/{build_id}/consoleText',
            _callback=self.jenkins._return_body,
        )

    def start(self,
              name: str,
              parameters: Optional[dict] = None,
              delay: Optional[int] = 0
              ) -> Optional[int]:
        """
        Enqueue new build with delay (default is 0 seconds, means immediately)

        Note about delay (quiet-period):
        https://www.jenkins.io/blog/2010/08/11/quiet-period-feature/

        Args:
            name (str):
                Job name or path (if in folder).

            parameters (int):
                Parameters of triggering build.

            delay (int):
                Delay before start.

        Returns:
            Optional[int]: queue item id.
        """
        def callback(response) -> Optional[int]:
            # FIXME: on Jenkins 1.554 there is problem, no queue id returned
            queue_item_url = response.headers['location']
            try:
                queue_id = queue_item_url.rstrip('/').split('/')[-1]
                return int(queue_id)
            except ValueError:
                return None

        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        path = f'/{folder_name}/job/{job_name}'

        data = None

        if parameters:
            formatted_parameters = [
                {'name': k, 'value': str(v)} for k, v in parameters.items()
            ]  # type: Any

            if len(formatted_parameters) == 1:
                formatted_parameters = formatted_parameters[0]

            data = {
                'json': json.dumps({
                    'parameter': formatted_parameters,
                    'statusCode': '303',
                    'redirectTo': '.',
                }),
                **parameters,
            }
            path += '/buildWithParameters'
        else:
            path += '/build'

        return self.jenkins._request(
            'POST',
            path,
            params={'delay': delay},
            data=data,
            _callback=callback,
        )

    def stop(self, name: str, build_id: Union[int, str]) -> None:
        """
        Stop specified build.

        Args:
            name (str):
                Job name or path (if in folder).

            build_id (int):
                Build number or some of standard tags like `lastBuild`.

        Returns:
            None
        """
        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        return self.jenkins._request(
            'POST',
            f'/{folder_name}/job/{job_name}/{build_id}/stop'
        )

    def delete(self, name: str, build_id: Union[int, str]) -> None:
        """
        Delete specified build.

        Args:
            name (str):
                Job name or path (if in folder).

            build_id (int):
                Build number or some of standard tags like `lastBuild`.

        Returns:
            None
        """
        folder_name, job_name = self.jenkins._get_folder_and_job_name(name)

        return self.jenkins._request(
            'POST',
            f'/{folder_name}/job/{job_name}/{build_id}/doDelete'
        )



import pytest

from ujenkins.exceptions import JenkinsError
from ujenkins.helpers import parse_build_url


def test_parse_build_urlt():
    name, number = parse_build_url(
        'http://localhost:8080/job/jobbb/1/console'
    )
    assert name == 'jobbb'
    assert number == 1

    name, number = parse_build_url(
        'http://localhost:8080/job/test_folder/job/test_job/567/console'
    )
    assert name == 'test_folder/test_job'
    assert number == 567

    with pytest.raises(JenkinsError):
        parse_build_url('xxx')



import re

import pytest
import responses

from ujenkins.exceptions import JenkinsError

NODES_JSON = """{
  "_class" : "hudson.model.ComputerSet",
  "busyExecutors" : 0,
  "computer" : [
    {
      "_class" : "hudson.model.Hudson$MasterComputer",
      "actions" : [

      ],
      "assignedLabels" : [
        {
          "name" : "master"
        }
      ],
      "description" : "the master Jenkins node",
      "displayName" : "master",
      "executors" : [
        {

        },
        {

        }
      ],
      "icon" : "computer-x.png",
      "iconClassName" : "icon-computer-x",
      "idle" : true,
      "jnlpAgent" : false,
      "launchSupported" : true,
      "loadStatistics" : {
        "_class" : "hudson.model.Label$1"
      },
      "manualLaunchAllowed" : true,
      "monitorData" : {
        "hudson.node_monitors.SwapSpaceMonitor" : {
          "_class" : "hudson.node_monitors.SwapSpaceMonitor$MemoryUsage2",
          "availablePhysicalMemory" : 78348288,
          "availableSwapSpace" : 1013350400,
          "totalPhysicalMemory" : 2082197504,
          "totalSwapSpace" : 1073737728
        },
        "hudson.node_monitors.TemporarySpaceMonitor" : {
          "_class" : "hudson.node_monitors.DiskSpaceMonitorDescriptor$DiskSpace",
          "timestamp" : 1632064529510,
          "path" : "/tmp",
          "size" : 502632448
        },
        "hudson.node_monitors.DiskSpaceMonitor" : {
          "_class" : "hudson.node_monitors.DiskSpaceMonitorDescriptor$DiskSpace",
          "timestamp" : 1632064529508,
          "path" : "/var/jenkins_home",
          "size" : 502632448
        },
        "hudson.node_monitors.ArchitectureMonitor" : "Linux (amd64)",
        "hudson.node_monitors.ResponseTimeMonitor" : {
          "_class" : "hudson.node_monitors.ResponseTimeMonitor$Data",
          "timestamp" : 1632064529535,
          "average" : 0
        },
        "hudson.node_monitors.ClockMonitor" : {
          "_class" : "hudson.util.ClockDifference",
          "diff" : 0
        }
      },
      "numExecutors" : 2,
      "offline" : true,
      "offlineCause" : {
        "_class" : "hudson.node_monitors.DiskSpaceMonitorDescriptor$DiskSpace"
      },
      "offlineCauseReason" : "Disk space is too low. Only 0.479GB left on /var/jenkins_home.",
      "oneOffExecutors" : [

      ],
      "temporarilyOffline" : true
    }
  ],
  "displayName" : "Nodes",
  "totalExecutors" : 0
}
"""

NODE_INFO_JSON = """{
  "_class" : "hudson.model.Hudson$MasterComputer",
  "actions" : [

  ],
  "assignedLabels" : [
    {
      "name" : "master"
    }
  ],
  "description" : "the master Jenkins node",
  "displayName" : "master",
  "executors" : [
    {

    },
    {

    }
  ],
  "icon" : "computer-x.png",
  "iconClassName" : "icon-computer-x",
  "idle" : true,
  "jnlpAgent" : false,
  "launchSupported" : true,
  "loadStatistics" : {
    "_class" : "hudson.model.Label$1"
  },
  "manualLaunchAllowed" : true,
  "monitorData" : {
    "hudson.node_monitors.SwapSpaceMonitor" : {
      "_class" : "hudson.node_monitors.SwapSpaceMonitor$MemoryUsage2",
      "availablePhysicalMemory" : 77758464,
      "availableSwapSpace" : 957181952,
      "totalPhysicalMemory" : 2082197504,
      "totalSwapSpace" : 1073737728
    },
    "hudson.node_monitors.TemporarySpaceMonitor" : {
      "_class" : "hudson.node_monitors.DiskSpaceMonitorDescriptor$DiskSpace",
      "timestamp" : 1632068934040,
      "path" : "/tmp",
      "size" : 502603776
    },
    "hudson.node_monitors.DiskSpaceMonitor" : {
      "_class" : "hudson.node_monitors.DiskSpaceMonitorDescriptor$DiskSpace",
      "timestamp" : 1632068934037,
      "path" : "/var/jenkins_home",
      "size" : 502603776
    },
    "hudson.node_monitors.ArchitectureMonitor" : "Linux (amd64)",
    "hudson.node_monitors.ResponseTimeMonitor" : {
      "_class" : "hudson.node_monitors.ResponseTimeMonitor$Data",
      "timestamp" : 1632068934043,
      "average" : 0
    },
    "hudson.node_monitors.ClockMonitor" : {
      "_class" : "hudson.util.ClockDifference",
      "diff" : 0
    }
  },
  "numExecutors" : 2,
  "offline" : true,
  "offlineCause" : {
    "_class" : "hudson.node_monitors.DiskSpaceMonitorDescriptor$DiskSpace"
  },
  "offlineCauseReason" : "Disk space is too low. Only 0.479GB left on /var/jenkins_home.",
  "oneOffExecutors" : [

  ],
  "temporarilyOffline" : true
}
"""

RSS_BUILDS_FAILED_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<feed xmlns="http://www.w3.org/2005/Atom">
    <title>Jenkins:master (failed builds)</title>
    <link rel="alternate" type="text/html" href="http://localhost:8080/computer/(master)/"></link>
    <updated>2022-06-06T21:14:09Z</updated>
    <author>
        <name>Jenkins Server</name>
    </author>
    <id>urn:uuid:903deee0-7bfa-11db-9fe1-0800200c9a66</id>
    <entry>
        <title>jobbb #15 (broken since this build)</title>
        <link rel="alternate" type="text/html" href="http://localhost:8080/job/jobbb/15/"></link>
        <id>tag:hudson.dev.java.net,2022:jobbb:15</id>
        <published>2022-06-06T21:14:09Z</published>
        <updated>2022-06-06T21:14:09Z</updated>
    </entry>
</feed>
"""

RSS_BUILDS_ALL_XML = r"""<?xml version="1.0" encoding="UTF-8"?>

<feed xmlns="http://www.w3.org/2005/Atom">
    <title>Jenkins:master (all builds)</title>
    <link rel="alternate" type="text/html" href="http://localhost:8080/computer/(master)/"></link>
    <updated>2022-06-06T21:14:09Z</updated>
    <author>
        <name>Jenkins Server</name>
    </author>
    <id>urn:uuid:903deee0-7bfa-11db-9fe1-0800200c9a66</id>
    <entry>
        <title>jobbb #15 (broken since this build)</title>
        <link rel="alternate" type="text/html" href="http://localhost:8080/job/jobbb/15/"></link>
        <id>tag:hudson.dev.java.net,2022:jobbb:15</id>
        <published>2022-06-06T21:14:09Z</published>
        <updated>2022-06-06T21:14:09Z</updated>
    </entry>
    <entry>
        <title>jobbb #14 (stable)</title>
        <link rel="alternate" type="text/html" href="http://localhost:8080/job/jobbb/14/"></link>
        <id>tag:hudson.dev.java.net,2021:jobbb:14</id>
        <published>2021-08-10T21:02:55Z</published>
        <updated>2021-08-10T21:02:55Z</updated>
    </entry>
    <entry>
        <title>jobbb #13 (stable)</title>
        <link rel="alternate" type="text/html" href="http://localhost:8080/job/jobbb/13/"></link>
        <id>tag:hudson.dev.java.net,2021:jobbb:13</id>
        <published>2021-08-10T21:02:06Z</published>
        <updated>2021-08-10T21:02:06Z</updated>
    </entry>
</feed>
"""

NODE_CONFIG_XML = r"""<?xml version="1.1" encoding="UTF-8"?>
<slave>
  <name>buildbot</name>
  <description></description>
  <remoteFS>/home/user/jenkins</remoteFS>
  <numExecutors>2</numExecutors>
  <mode>NORMAL</mode>
  <retentionStrategy class="hudson.slaves.RetentionStrategy$Always"/>
  <launcher class="hudson.slaves.JNLPLauncher">
    <workDirSettings>
      <disabled>false</disabled>
      <internalDir>remoting</internalDir>
      <failIfWorkDirIsMissing>false</failIfWorkDirIsMissing>
    </workDirSettings>
    <webSocket>false</webSocket>
  </launcher>
  <label></label>
  <nodeProperties/>
"""


@responses.activate
def test_get(client):
    responses.add(
        responses.GET,
        re.compile(r'.+/computer/api/json'),
        content_type='application/json;charset=utf-8',
        body=NODES_JSON,
    )

    response = client.nodes.get()
    assert len(response) == 1


@responses.activate
def test_get_failed_builds(client):
    responses.add(
        responses.GET,
        re.compile(r'.+/computer/.+/rssFailed'),
        content_type='application/atom+xml;charset=UTF-8',
        body=RSS_BUILDS_FAILED_XML,
    )

    response = client.nodes.get_failed_builds('master')
    assert len(response) == 1
    assert response[0]['url'] == 'http://localhost:8080/job/jobbb/15/'
    assert response[0]['job_name'] == 'jobbb'
    assert response[0]['number'] == 15


@responses.activate
def test_get_all_builds(client):
    responses.add(
        responses.GET,
        re.compile(r'.+/computer/.+/rssAll'),
        content_type='application/atom+xml;charset=UTF-8',
        body=RSS_BUILDS_ALL_XML,
    )

    response = client.nodes.get_all_builds('master')
    assert len(response) == 3


@responses.activate
def test_get_info(client):
    responses.add(
        responses.GET,
        re.compile(r'.+/computer/.+/api/json'),
        content_type='application/json;charset=utf-8',
        body=NODE_INFO_JSON,
    )

    response = client.nodes.get_info('master')
    assert response['numExecutors'] == 2


@responses.activate
def test_get_config(client):
    responses.add(
        responses.GET,
        re.compile(r'.+/computer/.+/config.xml'),
        content_type='application/xml',
        body=NODE_CONFIG_XML,
    )

    response = client.nodes.get_config('buildbot')
    assert '<name>buildbot</name>' in response


@responses.activate
def test_is_exists(client):
    responses.add(
        responses.GET,
        re.compile(r'.+/computer/.+/api/json'),
        status=404,
    )

    assert client.nodes.is_exists('master') is False


@responses.activate
def test_create(client):
    responses.add(
        responses.GET,
        re.compile(r'.+/computer/api/json'),
        content_type='application/json;charset=utf-8',
        body=NODES_JSON,
    )

    responses.add(
        responses.POST,
        re.compile(r'.+/computer/doCreateItem'),
    )

    assert client.nodes.create('buildbot', {}) is None


@responses.activate
def test_delete(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/computer/.+/doDelete'),
    )

    assert client.nodes.delete('buildbot') is None


@responses.activate
def test_reconfigure(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/computer/.+/config.xml'),
    )

    assert client.nodes.reconfigure('buildbot', NODE_CONFIG_XML) is None

    with pytest.raises(JenkinsError):
        client.nodes.reconfigure('(master)', NODE_CONFIG_XML)


@pytest.mark.asyncio
async def test_async_enable(aiohttp_mock, async_client):
    aiohttp_mock.get(
        re.compile(r'.+/computer/.+/api/json'),
        content_type='application/json;charset=utf-8',
        body=NODE_INFO_JSON,
    )

    aiohttp_mock.post(
        re.compile(r'.+/computer/.+/toggleOffline'),
        content_type='text/plain',
        body='',
    )

    response = await async_client.nodes.enable('master')
    assert response is None


@responses.activate
def test_enable_when_disabled(client):
    responses.add(
        responses.GET,
        re.compile(r'.+/computer/.+/api/json'),
        content_type='application/json;charset=utf-8',
        body='{"offline": true, "temporarilyOffline": true}',
    )

    responses.add(
        responses.POST,
        re.compile(r'.+/computer/.+/toggleOffline'),
        content_type='text/plain',
    )

    assert client.nodes.enable('master') is None
    assert len(responses.calls) == 2


@responses.activate
def test_disable_when_enabled(client):
    responses.add(
        responses.GET,
        re.compile(r'.+/computer/.+/api/json'),
        content_type='application/json;charset=utf-8',
        body='{"offline": false, "temporarilyOffline": false}',
    )

    responses.add(
        responses.POST,
        re.compile(r'.+/computer/.+/toggleOffline'),
        content_type='text/plain',
    )

    assert client.nodes.disable('master') is None
    assert len(responses.calls) == 2


@responses.activate
def test_disable_when_disabled(client):
    responses.add(
        responses.GET,
        re.compile(r'.+/computer/.+/api/json'),
        content_type='application/json;charset=utf-8',
        body='{"offline": true, "temporarilyOffline": true}',
    )

    responses.add(
        responses.POST,
        re.compile(r'.+/computer/.+/toggleOffline'),
        content_type='text/plain',
    )

    assert client.nodes.disable('master') is None
    assert len(responses.calls) == 1


@responses.activate
def test_update_offline_reason(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/computer/.+/changeOfflineCause'),
    )

    assert client.nodes.update_offline_reason('buildbot', 'maintenance') is None


@responses.activate
def test_launch_agent(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/computer/.+/launchSlaveAgent'),
    )

    assert client.nodes.launch_agent('buildbot') is None



JENKINS_INFO_JSON = """
{
    "_class": "hudson.model.Hudson",
    "assignedLabels": [
        {
            "name": "built-in"
        }
    ],
    "description": null,
    "jobs": [
        {
            "_class" : "hudson.model.FreeStyleProject",
            "name" : "test",
            "url" : "http://localhost:8080/job/test/",
            "color" : "red"
        },
        {
            "_class" : "com.cloudbees.hudson.plugins.folder.Folder",
            "name" : "test_folder",
            "url" : "http://localhost:8080/job/test_folder/"
        }
    ],
    "mode": "NORMAL",
    "nodeDescription": "the Jenkins controller's built-in node",
    "nodeName": "",
    "numExecutors": 2,
    "overallLoad": {},
    "primaryView": {
        "_class": "hudson.model.AllView",
        "name": "all",
        "url": "http://localhost:8080/"
    },
    "quietDownReason": null,
    "quietingDown": false,
    "slaveAgentPort": 50000,
    "unlabeledLoad": {
        "_class": "jenkins.model.UnlabeledLoadStatistics"
    },
    "url": "http://localhost:8080/",
    "useCrumbs": true,
    "useSecurity": true,
    "views": [
        {
            "_class": "hudson.model.AllView",
            "name": "all",
            "url": "http://localhost:8080/"
        },
        {
            "_class": "hudson.model.MyView",
            "name": "another",
            "url": "http://localhost:8080/view/another/"
        },
        {
            "_class": "hudson.model.ListView",
            "name": "my_view",
            "url": "http://localhost:8080/view/my_view/"
        }
    ]
}
"""



import re

import responses

from tests import JENKINS_INFO_JSON

TOKEN_JSON = """
  {
    "status": "ok",
    "data": {
        "tokenName": "test",
        "tokenUuid": "a8f93d3d-e431-4c88-ad6f-f71dd50e774f",
        "tokenValue":"11064db87a080c7dc478fedfd8cd9d4265"
    }
  }
"""


@responses.activate
def test_get_status(client):
    responses.add(
        responses.GET,
        re.compile(r'.+/api/json'),
        content_type='application/json;charset=utf-8',
        body=JENKINS_INFO_JSON
    )

    status = client.system.get_status()
    assert status['quietingDown'] is False


@responses.activate
def test_get_version(client):
    responses.add(
        responses.GET,
        re.compile(r'.+/'),
        headers={'X-Jenkins': '2.0.129'}
    )

    version = client.system.get_version()
    assert version.major == 2
    assert version.minor == 0
    assert version.patch == 129

    responses.add(
        responses.GET,
        re.compile(r'.+/'),
        headers={'X-Jenkins': '2.123'}
    )

    version = client.system.get_version()
    assert version.major == 2
    assert version.minor == 123
    assert version.patch == 0


@responses.activate
def test_is_ready(client):
    responses.add(
        responses.GET,
        re.compile(r'.+/api/json'),
        content_type='application/json;charset=utf-8',
        body=JENKINS_INFO_JSON
    )

    assert client.system.is_ready() is True


@responses.activate
def test_quiet_down(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/quietDown'),
    )

    assert client.system.quiet_down() is None


@responses.activate
def test_cancel_quiet_down(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/cancelQuietDown'),
    )

    assert client.system.cancel_quiet_down() is None


@responses.activate
def test_restart(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/restart'),
    )

    assert client.system.restart() is None


@responses.activate
def test_safe_restart(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/safeRestart'),
    )

    assert client.system.safe_restart() is None


@responses.activate
def test_generate_token(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/me/descriptorByName/jenkins.security.ApiTokenProperty/generateNewToken'),
        content_type='application/json;charset=utf-8',
        body=TOKEN_JSON,
    )

    value, uuid = client.system.generate_token('test')
    assert uuid == 'a8f93d3d-e431-4c88-ad6f-f71dd50e774f'
    assert value == '11064db87a080c7dc478fedfd8cd9d4265'


@responses.activate
def test_revoke_token(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/me/descriptorByName/jenkins.security.ApiTokenProperty/revoke'),
    )

    assert client.system.revoke_token('a8f93d3d-e431-4c88-ad6f-f71dd50e774f') is None


@responses.activate
def test_run_groovy_script(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/scriptText'),
        body='test'
    )

    response = client.system.run_groovy_script('print("test")')
    assert response == 'test'



import pytest

from aioresponses import aioresponses

from ujenkins import AsyncJenkinsClient, JenkinsClient


@pytest.fixture
def aiohttp_mock():
    with aioresponses() as mock:
        yield mock


@pytest.fixture
def client():
    j = JenkinsClient('http://server')
    # disable crumb wrapper for simplify testing
    j.crumb = False
    yield j
    j.close()


@pytest.fixture
async def async_client():
    j = AsyncJenkinsClient('http://server')
    # disable crumb wrapper for simplify testing
    j.crumb = False
    yield j
    await j.close()



import re

from http import HTTPStatus

import responses

from tests import JENKINS_INFO_JSON

JOB_INFO_JSON = """
{
  "_class" : "hudson.model.FreeStyleProject",
  "actions" : [
    {

    },
    {

    }
  ],
  "description" : "",
  "displayName" : "test",
  "displayNameOrNull" : null,
  "fullDisplayName" : "test",
  "fullName" : "test",
  "name" : "test",
  "url" : "http://localhost:8080/job/test/",
  "buildable" : true,
  "builds" : [
    {
      "_class" : "hudson.model.FreeStyleBuild",
      "number" : 15,
      "url" : "http://localhost:8080/job/test/15/"
    },
    {
      "_class" : "hudson.model.FreeStyleBuild",
      "number" : 14,
      "url" : "http://localhost:8080/job/test/14/"
    },
    {
      "_class" : "hudson.model.FreeStyleBuild",
      "number" : 13,
      "url" : "http://localhost:8080/job/test/13/"
    },
    {
      "_class" : "hudson.model.FreeStyleBuild",
      "number" : 12,
      "url" : "http://localhost:8080/job/test/12/"
    }
  ],
  "color" : "red",
  "firstBuild" : {
    "_class" : "hudson.model.FreeStyleBuild",
    "number" : 1,
    "url" : "http://localhost:8080/job/test/1/"
  },
  "healthReport" : [
    {
      "description" : "Build stability: 1 out of the last 5 builds failed.",
      "iconClassName" : "icon-health-60to79",
      "iconUrl" : "health-60to79.png",
      "score" : 80
    }
  ],
  "inQueue" : false,
  "keepDependencies" : false,
  "lastBuild" : {
    "_class" : "hudson.model.FreeStyleBuild",
    "number" : 15,
    "url" : "http://localhost:8080/job/test/15/"
  },
  "lastCompletedBuild" : {
    "_class" : "hudson.model.FreeStyleBuild",
    "number" : 15,
    "url" : "http://localhost:8080/job/test/15/"
  },
  "lastFailedBuild" : {
    "_class" : "hudson.model.FreeStyleBuild",
    "number" : 15,
    "url" : "http://localhost:8080/job/test/15/"
  },
  "lastStableBuild" : {
    "_class" : "hudson.model.FreeStyleBuild",
    "number" : 14,
    "url" : "http://localhost:8080/job/test/14/"
  },
  "lastSuccessfulBuild" : {
    "_class" : "hudson.model.FreeStyleBuild",
    "number" : 14,
    "url" : "http://localhost:8080/job/test/14/"
  },
  "lastUnstableBuild" : null,
  "lastUnsuccessfulBuild" : {
    "_class" : "hudson.model.FreeStyleBuild",
    "number" : 15,
    "url" : "http://localhost:8080/job/test/15/"
  },
  "nextBuildNumber" : 16,
  "property" : [

  ],
  "queueItem" : null,
  "concurrentBuild" : false,
  "disabled" : false,
  "downstreamProjects" : [

  ],
  "labelExpression" : null,
  "scm" : {
    "_class" : "hudson.scm.NullSCM"
  },
  "upstreamProjects" : [

  ]
}
"""

JOB_CONFIG_XML = """
<?xml version='1.1' encoding='UTF-8'?>
<project>
  <description></description>
  <keepDependencies>false</keepDependencies>
  <properties/>
  <scm class="hudson.scm.NullSCM"/>
  <canRoam>true</canRoam>
  <disabled>false</disabled>
  <blockBuildWhenDownstreamBuilding>false</blockBuildWhenDownstreamBuilding>
  <blockBuildWhenUpstreamBuilding>false</blockBuildWhenUpstreamBuilding>
  <triggers/>
  <concurrentBuild>false</concurrentBuild>
  <builders/>
  <publishers/>
  <buildWrappers/>
</project>
"""


@responses.activate
def test_get(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/api/json'),
        body=JENKINS_INFO_JSON,
    )

    response = client.jobs.get()
    assert len(response) == 2
    assert 'test' in response


@responses.activate
def test_get_info(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/test/api/json'),
        content_type='application/json;charset=utf-8',
        body=JOB_INFO_JSON,
    )

    response = client.jobs.get_info('test')
    assert len(response['builds']) == 4
    assert response['buildable'] is True


@responses.activate
def test_get_config(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/test/config.xml'),
        content_type='application/xml',
        body=JOB_CONFIG_XML,
    )

    response = client.jobs.get_config('test')
    assert '<description></description>' in response


@responses.activate
def test_is_exists(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/test/api/json'),
        content_type='application/json;charset=utf-8',
        body=JOB_INFO_JSON,
    )

    responses.add(
        responses.GET,
        re.compile(r'.*/another_job/api/json'),
        status=HTTPStatus.NOT_FOUND,
    )

    assert client.jobs.is_exists('test') is True
    assert client.jobs.is_exists('another_job') is False


@responses.activate
def test_create(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/createItem'),
    )

    assert client.jobs.create('some_job', JOB_CONFIG_XML) is None


@responses.activate
def test_reconfigure(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/some_job/config.xml'),
    )

    assert client.jobs.reconfigure('some_job', JOB_CONFIG_XML) is None


@responses.activate
def test_delete(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/job/.+/doDelete'),
    )

    assert client.jobs.delete('useless_job') is None


@responses.activate
def test_copy(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/createItem'),
        match=[responses.matchers.query_param_matcher({
            'mode': 'copy',
            'from': 'job_old',
            'name': 'job_new',
        })],
    )

    assert client.jobs.copy('job_old', 'job_new') is None


@responses.activate
def test_rename(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/doRename'),
        match=[responses.matchers.query_param_matcher({
            'newName': 'job_new',
        })],
    )

    assert client.jobs.rename('job_old', 'job_new') is None


@responses.activate
def test_enable(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/enable'),
    )

    assert client.jobs.enable('job') is None


@responses.activate
def test_disable(client):
    responses.add(
        responses.POST,
        re.compile(r'.+/disable'),
    )

    assert client.jobs.disable('job') is None



from _typeshed import Incomplete

class JenkinsError(Exception):
    message: Incomplete
    status: Incomplete
    def __init__(self, message: Incomplete | None = ..., status: Incomplete | None = ...) -> None: ...

class JenkinsNotFoundError(JenkinsError): ...



import json
import xml.etree.ElementTree

from functools import partial
from typing import Any, Dict, List, Optional

from ujenkins.exceptions import JenkinsError, JenkinsNotFoundError
from ujenkins.helpers import parse_build_url


def _parse_rss(rss: str) -> List[dict]:
    builds = []
    ns = {'atom': 'http://www.w3.org/2005/Atom'}

    root = xml.etree.ElementTree.fromstring(rss)
    for entry in root.findall('atom:entry', ns):
        link = entry.find('atom:link', ns)
        if link is not None:
            build_url = link.attrib['href']
            job_name, build_id = parse_build_url(build_url)

            builds.append({
                'url': build_url,
                'job_name': job_name,
                'number': build_id,
            })

    return list(reversed(builds))


class Nodes:

    def __init__(self, jenkins):
        self.jenkins = jenkins

    @staticmethod
    def _normalize_name(name: str) -> str:
        # embedded node `master` actually have brackets in HTTP requests
        if name in ('master', 'Built-In Node'):
            return '(master)'
        return name

    def get(self) -> Dict[str, dict]:
        """
        Get all available nodes on server.

        Returns:
            Dict[str, dict]: node name, and it`s detailed information.

        Example:

        .. code-block:: python

            response = {
                "master": dict(...),
                "buildbot1": dict(...)
            }

        """
        def callback(response):
            nodes = json.loads(response.body)
            return {v['displayName']: v for v in nodes['computer']}

        return self.jenkins._request(
            'GET',
            '/computer/api/json',
            _callback=callback,
        )

    def get_failed_builds(self, name: str) -> List[dict]:
        """
        Return list of detalizied failed builds for node name. Actually it
        parsed from RSS feed. usefull for build restart. Ascending builds sort.

        Args:
            name (str):
                Node name.

        Returns:
            List[dict]: builds and their information.

        Example:

        .. code-block:: python

            response = [{
                'job_name': 'test',
                'number': 1,
                'url': 'http://localhost:8080/job/test/1/'
            }]
        """
        def callback(response) -> List[dict]:
            return _parse_rss(response.body)

        name = self._normalize_name(name)

        return self.jenkins._request(
            'GET',
            f'/computer/{name}/rssFailed',
            _callback=callback
        )

    def get_all_builds(self, name: str) -> List[dict]:
        """
        Return list of all detalizied builds for node name, actually it parsed
        from RSS feed. Ascending builds sort.

        Args:
            name (str):
                Node name.

        Returns:
            List[dict]: list of all builds for specified node.

        Example:

        .. code-block:: python

            response = [{
                'job_name': 'test',
                'number': 1,
                'url': 'http://localhost:8080/job/test/1/'
            }]
        """
        def callback(response) -> List[dict]:
            return _parse_rss(response.body)

        name = self._normalize_name(name)

        return self.jenkins._request(
            'GET',
            f'/computer/{name}/rssAll',
            _callback=callback
        )

    def get_info(self, name: str) -> dict:
        """
        Get node detailed information.

        Args:
            name (str):
                Node name.

        Returns:
            dict: detailed node information.
        """
        def callback1(_):
            return self.jenkins._request('GET', f'/computer/{name}/api/json')

        def callback2(response: Any):
            if isinstance(response, JenkinsError):
                return response

            response['_disconnected'] = (
                response['offline'] is True and
                response['temporarilyOffline'] is False
            )

            return response

        name = self._normalize_name(name)
        return self.jenkins._chain([callback1, callback2])

    def get_config(self, name: str) -> str:
        """
        Return node config in XML format.

        Args:
            name (str):
                Node name.

        Returns:
            str: node config.
        """
        name = self._normalize_name(name)

        return self.jenkins._request(
            'GET',
            f'/computer/{name}/config.xml',
            _callback=self.jenkins._return_body,
        )

    def is_exists(self, name: str) -> bool:
        """
        Check is node exist.

        Args:
            name (str):
                Node name.

        Returns:
            bool: node existing.
        """
        if name == '':
            return False

        def callback1(_) -> Any:
            return partial(self.get_info, name)

        def callback2(response: Any) -> bool:
            if isinstance(response, JenkinsNotFoundError):
                return False

            return True

        return self.jenkins._chain([callback1, callback2])

    def create(self, name: str, config: dict) -> None:
        """
        Create new node.

        Args:
            name (str):
                Node name.

            config (str):
                XML config for new node.

        Returns:
            None

        Raises:
            JenkinsError: in case node already exists.
        """
        def callback1(_) -> Any:
            return partial(self.get)

        def callback2(response: Any) -> bool:
            # if not check, then we get 400 error with unclear stacktrace
            if name in response:
                raise JenkinsError(f'Node `{name}` is already exists')

            if 'type' not in config:
                config['type'] = 'hudson.slaves.DumbSlave'

            config['name'] = name

            params = {
                'name': name,
                'type': config['type'],
                'json': json.dumps(config)
            }

            return self.jenkins._request(
                'POST',
                '/computer/doCreateItem',
                params=params,
            )

        return self.jenkins._chain([callback1, callback2])

    def delete(self, name: str) -> None:
        """
        Delete node.

        Args:
            name (str):
                Node name.

        Returns:
            None
        """
        name = self._normalize_name(name)

        return self.jenkins._request(
            'POST',
            f'/computer/{name}/doDelete'
        )

    def reconfigure(self, name: str, config: str) -> None:
        """
        Reconfigure node.

        Args:
            name (str):
                Node name.

            config (str):
                New XML config for node.

        Returns:
            None
        """
        name = self._normalize_name(name)

        if name == '(master)':
            raise JenkinsError('Cannot reconfigure master node')

        return self.jenkins._request(
            'POST',
            f'/computer/{name}/config.xml',
            data=config,
            headers={'Content-Type': 'text/xml'},
        )

    def enable(self, name: str) -> None:
        """
        Enable node if it disabled.

        Args:
            name (str):
                Node name.

        Returns:
            None
        """
        name = self._normalize_name(name)

        def callback1(_) -> Any:
            return partial(self.get_info, name)

        def callback2(response: dict) -> None:
            # skip if already enabled
            if response['temporarilyOffline'] is False:
                return None

            return self.jenkins._request('POST', f'/computer/{name}/toggleOffline')

        return self.jenkins._chain([callback1, callback2])

    def disable(self, name: str, message: Optional[str] = '') -> None:
        """
        Disable node if it enabled.

        Args:
            name (str):
                Node name.

            message (Optional[str]):
                Reason message.

        Returns:
            None
        """
        name = self._normalize_name(name)

        def callback1(_) -> Any:
            return partial(self.get_info, name)

        def callback2(response: dict) -> None:
            # skip if already disabled
            if response['temporarilyOffline'] is True:
                return None

            return self.jenkins._request(
                'POST',
                f'/computer/{name}/toggleOffline',
                params={'offlineMessage': message},
            )

        return self.jenkins._chain([callback1, callback2])

    def update_offline_reason(self, name: str, message: str) -> None:
        """
        Update reason message of disabled node.

        Args:
            name (str):
                Node name.

            message (str):
                Reason message.

        Returns:
            None
        """
        name = self._normalize_name(name)

        return self.jenkins._request(
            'POST',
            f'/computer/{name}/changeOfflineCause',
            params={'offlineMessage': message}
        )

    def launch_agent(self, name: str) -> None:
        """
        Launch agent on node, for example in case when disconnected.

        State of connection can be determinated by `get_info(...)` method,
        which contains custom property defined by packages: `_disconnected`.

        Args:
            name (str):
                Node name.

        Returns:
            None
        """
        name = self._normalize_name(name)

        return self.jenkins._request(
            'POST',
            f'/computer/{name}/launchSlaveAgent',
        )



from ujenkins.adapters.aio import AsyncJenkinsClient as AsyncJenkinsClient
from ujenkins.adapters.sync import JenkinsClient as JenkinsClient
from ujenkins.exceptions import JenkinsError as JenkinsError, JenkinsNotFoundError as JenkinsNotFoundError



import json

from typing import NamedTuple, Tuple

from ujenkins.exceptions import JenkinsError


class JenkinsVersion(NamedTuple):
    major: int
    minor: int
    patch: int = 0
    build: int = 0


class System:

    def __init__(self, jenkins):
        self.jenkins = jenkins

    def get_status(self) -> dict:
        """
        Get server status.

        Returns:
            dict: jenkins server details.
        """
        return self.jenkins._request('GET', '/api/json')

    def get_version(self) -> JenkinsVersion:
        """
        Get server version.

        Returns:
            JenkinsVersion: named tuple with minor, major, patch version.
        """
        def callback(response):
            header = response.headers.get('X-Jenkins')
            if not header:
                raise JenkinsError('Header `X-Jenkins` isn`t found in response')

            versions = header.split('.')

            return JenkinsVersion(*map(int, versions))

        return self.jenkins._request('GET', '/', _callback=callback)

    def is_ready(self) -> bool:
        """
        Determines is server loaded and ready for work.

        Returns:
            bool: ready state.
        """
        def callback(response):
            try:
                status = response.body
                return 'mode' in status
            except JenkinsError:
                return False

        return self.jenkins._request('GET', '/api/json', _callback=callback)

    def quiet_down(self) -> None:
        """
        Start server quiet down period, new builds will not be started.

        Returns:
            None
        """
        return self.jenkins._request('POST', '/quietDown')

    def cancel_quiet_down(self) -> None:
        """
        Cancel server quiet down period.

        Returns:
            None
        """
        return self.jenkins._request('POST', '/cancelQuietDown')

    def restart(self) -> None:
        """
        Restart server immediately.

        Returns:
            None
        """
        return self.jenkins._request('POST', '/restart')

    def safe_restart(self) -> None:
        """
        Restart server when installation is complete and no jobs are running.

        Returns:
            None
        """
        return self.jenkins._request('POST', '/safeRestart')

    @staticmethod
    def _build_token_url(suffix: str) -> str:
        return '/me/descriptorByName/jenkins.security.ApiTokenProperty/' + suffix

    def generate_token(self, name: str) -> Tuple[str, str]:
        """
        Generate new API token.

        Args:
            name (str):
                Name of token.

        Returns:
            Tuple[str, str]: tokenValue - uses for authorization,
                             tokenUuid - uses for revoke
        """
        def callback(response):
            content = json.loads(response.body)

            if content['status'] != 'ok':
                raise JenkinsError('Non OK status returned: ' + str(content))

            return content['data']['tokenValue'], content['data']['tokenUuid']

        params = {'newTokenName': name}

        return self.jenkins._request(
            'POST',
            self._build_token_url('generateNewToken'),
            params=params,
            _callback=callback,
        )

    def revoke_token(self, token_uuid: str) -> None:
        """
        Revoke API token, please note that uuid is used, not value.

        Args:
            token_uuid (str):
                UUID of token to be revoked.

        Returns:
            None
        """
        params = {'tokenUuid': token_uuid}

        return self.jenkins._request(
            'POST',
            self._build_token_url('revoke'),
            params=params
        )

    def run_groovy_script(self, script: str) -> str:
        """
        Execute Groovy script on the server.

        Args:
            script (str):
                Script content.

        Returns:
            str: output of script.
        """
        return self.jenkins._request(
            'POST',
            '/scriptText',
            data={'script': script},
            _callback=self.jenkins._return_body,
        )



import re

import responses

BUILDS_ALL_JSON = """{
    "_class": "hudson.model.FreeStyleProject",
    "allBuilds": [
        {
            "_class": "hudson.model.FreeStyleBuild",
            "number": 1,
            "url": "http://localhost:8080/job/jobbb/1/"
        }
    ]
}
"""

BUILD_INFO_JSON = """{
  "_class" : "hudson.model.FreeStyleBuild",
  "actions" : [
    {
      "_class" : "hudson.model.CauseAction",
      "causes" : [
        {
          "_class" : "hudson.model.Cause$UserIdCause",
          "shortDescription" : "Started by user admin",
          "userId" : "admin",
          "userName" : "admin"
        }
      ]
    }
  ],
  "artifacts" : [

  ],
  "building" : false,
  "description" : null,
  "displayName" : "#14",
  "duration" : 6,
  "estimatedDuration" : 4,
  "executor" : null,
  "fullDisplayName" : "jobbb #14",
  "id" : "14",
  "keepLog" : false,
  "number" : 14,
  "queueId" : 429,
  "result" : "SUCCESS",
  "timestamp" : 1628629375963,
  "url" : "http://localhost:8080/job/jobbb/14/",
  "builtOn" : "",
  "changeSet" : {
    "_class" : "hudson.scm.EmptyChangeLogSet",
    "items" : [

    ],
    "kind" : null
  },
  "culprits" : [

  ]
}
"""


@responses.activate
def test_get(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/api/json'),
        body=BUILDS_ALL_JSON,
    )

    response = client.builds.get('job')
    assert len(response) == 1
    assert response[0]['number'] == 1


@responses.activate
def test_get_info(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/api/json'),
        content_type='application/json;charset=utf-8',
        body=BUILD_INFO_JSON,
    )

    response = client.builds.get_info('jobbb', 14)
    assert response['duration'] == 6


@responses.activate
def test_get_output(client):
    responses.add(
        responses.GET,
        re.compile(r'.*/job/.+/consoleText'),
        content_type='text/plain;charset=utf-8',
        body='Started by user admin\nRunning as SYSTEM',
    )

    response = client.builds.get_output('job', 14)
    assert 'Started' in response


@responses.activate
def test_start(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/job/.+/build.+'),
        headers={'Location': 'http://localhost:8080/queue/item/424/'}
    )

    response = client.builds.start('job')
    assert response == 424


@responses.activate
def test_stop(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/job/.+/stop'),
    )

    assert client.builds.stop('job', 424) is None


@responses.activate
def test_delete(client):
    responses.add(
        responses.POST,
        re.compile(r'.*/job/.+/doDelete'),
    )

    assert client.builds.delete('job', 424) is None



from ujenkins.adapters.aio import AsyncJenkinsClient
from ujenkins.adapters.sync import JenkinsClient
from ujenkins.exceptions import JenkinsError, JenkinsNotFoundError

__version__ = '0.5.0'

__all__ = (
    # adapters
    'AsyncJenkinsClient',
    'JenkinsClient',
    # exceptions
    'JenkinsError',
    'JenkinsNotFoundError',
)



import re
import sys

from http import HTTPStatus

import aiohttp
import pytest
import responses

from ujenkins import AsyncJenkinsClient, JenkinsClient, JenkinsError

CRUMB_JSON = """
{
  "_class": "hudson.security.csrf.DefaultCrumbIssuer",
  "crumb": "9c427004e7ed327a230436ee3103856d8df1eec7f2964a87d3d95e850974c4cd",
  "crumbRequestField": "Jenkins-Crumb"
}
"""


@responses.activate
def test_sync_client_retry():
    # responses library does`t support Retry mock
    # https://github.com/getsentry/responses/issues/135
    # so, just cover code of retry constructor
    statuses = [HTTPStatus.BAD_REQUEST, HTTPStatus.INTERNAL_SERVER_ERROR]

    client = JenkinsClient(
        'http://server',
        'user',
        'password',
        retry=dict(
            total=10,
            factor=1,
            statuses=statuses,
        )
    )

    assert client.session.adapters['http://'].max_retries.status_forcelist == statuses


@pytest.mark.asyncio
async def test_async_client_retry(aiohttp_mock):
    client = AsyncJenkinsClient(
        'http://server',
        'user',
        'password',
        retry=dict(
            total=10,
            statuses=[HTTPStatus.INTERNAL_SERVER_ERROR],
        )
    )
    client.crumb = False

    aiohttp_mock.get(
        'http://server/',
        content_type='text/plain',
        body='Server error',
        status=HTTPStatus.INTERNAL_SERVER_ERROR,
    )

    aiohttp_mock.get(
        'http://server/',
        content_type='text/plain',
        headers={'X-Jenkins': '2.0.129'},
        status=HTTPStatus.OK,
    )

    version = await client.system.get_version()
    assert version.major == 2
    assert version.minor == 0
    assert version.patch == 129
    await client.close()


@pytest.mark.asyncio
async def test_async_client_retry_exception(aiohttp_mock):
    client = AsyncJenkinsClient(
        'http://server',
        'user',
        'password',
        retry=dict(
            total=2,
            statuses=[HTTPStatus.INTERNAL_SERVER_ERROR],
        )
    )

    aiohttp_mock.get(
        'http://server/rest/version',
        exception=aiohttp.ClientError()
    )

    aiohttp_mock.get(
        'http://server/rest/version',
        exception=aiohttp.ClientError()
    )

    with pytest.raises(JenkinsError):
        await client.system.get_version()

    await client.close()


def test_retry_argument_validation():
    with pytest.raises(JenkinsError):
        JenkinsClient('http://server', retry=dict(total=1, strange_argument=1))

    with pytest.raises(JenkinsError):
        AsyncJenkinsClient('http://server', retry=dict(total=0))


@responses.activate
def test_sync_crumb():
    responses.add(
        responses.GET,
        re.compile(r'.+/crumbIssuer/api/json'),
        content_type='application/json;charset=utf-8',
        body=CRUMB_JSON
    )

    responses.add(
        responses.GET,
        re.compile(r'.+/'),
        headers={'X-Jenkins': '2.0.129'}
    )

    client = JenkinsClient('http://server')
    version = client.system.get_version()

    assert client.crumb['Jenkins-Crumb'] == \
        '9c427004e7ed327a230436ee3103856d8df1eec7f2964a87d3d95e850974c4cd'
    assert version.major == 2


@responses.activate
def test_sync_crumb_disabled():
    responses.add(
        responses.GET,
        re.compile(r'.+/crumbIssuer/api/json'),
        status=HTTPStatus.NOT_FOUND
    )

    responses.add(
        responses.GET,
        re.compile(r'.+/'),
        headers={'X-Jenkins': '2.0.129'}
    )

    client = JenkinsClient('http://server')
    version = client.system.get_version()

    assert client.crumb is False
    assert version.major == 2


@pytest.mark.skipif(sys.version_info < (3, 6), reason='aiohttp mock problem')
@pytest.mark.asyncio
async def test_async_crumb(aiohttp_mock, async_client):
    aiohttp_mock.get(
        re.compile(r'.+/crumbIssuer/api/json'),
        content_type='application/json;charset=utf-8',
        body=CRUMB_JSON
    )

    aiohttp_mock.get(
        re.compile(r'.+/'),
        headers={'X-Jenkins': '2.0.129'}
    )

    async_client.crumb = None
    version = await async_client.system.get_version()

    assert async_client.crumb['Jenkins-Crumb'] == \
        '9c427004e7ed327a230436ee3103856d8df1eec7f2964a87d3d95e850974c4cd'
    assert version.major == 2



import json

from collections import namedtuple
from http import HTTPStatus
from typing import Any, Callable, Optional, Tuple

from ujenkins.endpoints import Builds, Jobs, Nodes, System, Views
from ujenkins.exceptions import JenkinsError, JenkinsNotFoundError

Response = namedtuple('Response', ['status', 'headers', 'body'])


class Jenkins:

    def __init__(self):
        self.builds = Builds(self)
        self.jobs = Jobs(self)
        self.nodes = Nodes(self)
        self.system = System(self)
        self.views = Views(self)

    @staticmethod
    def _process(response: Response, callback: Optional[Callable] = None) -> Any:

        if response.status == HTTPStatus.NOT_FOUND:
            raise JenkinsNotFoundError(response.body)

        if response.status >= HTTPStatus.BAD_REQUEST:
            if response.status in (
                    HTTPStatus.UNAUTHORIZED,
                    HTTPStatus.FORBIDDEN,
                    HTTPStatus.INTERNAL_SERVER_ERROR):
                details = 'probably authentication problem:\n' + response.body
            else:
                details = '\n' + response.body

            raise JenkinsError(
                f'Request error [{response.status}], {details}',
                status=response.status,
            )

        # TODO: add response type annotations, parse json for callback
        if callback:
            return callback(response)

        if 'application/json' in response.headers.get('Content-Type', ''):
            return json.loads(response.body)

        return None

    @staticmethod
    def _get_folder_and_job_name(name: str) -> Tuple[str, str]:
        parts = name.split('/')

        job_name = parts[-1]
        folder_name = ''

        for folder in parts[:-1]:
            folder_name += f'job/{folder}/'

        return folder_name, job_name

    @staticmethod
    def _validate_retry_argument(retry: dict) -> None:
        for key in retry:
            if key not in ('total', 'factor', 'statuses'):
                raise JenkinsError('Unknown key in retry argument: ' + key)

        if retry.get('total', 0) <= 0:
            raise JenkinsError('Invalid `total` in retry argument must be > 0')

    @staticmethod
    def _return_body(response: Response) -> str:
        return response.body



from _typeshed import Incomplete
from typing import List, Optional, Tuple
from ujenkins.exceptions import JenkinsError as JenkinsError

JOB_BUILD_URL_RE: Incomplete

def construct_job_config(*, description: Optional[str] = ..., parameters: Optional[List[dict]] = ..., commands: Optional[List[str]] = ...) -> str: ...
def construct_node_config(*, name: str, remote_fs: Optional[str] = ..., executors: Optional[int] = ...) -> dict: ...
def parse_build_url(build_url: str) -> Tuple[str, int]: ...



import os
import re

from setuptools import setup, find_packages

init_file_path = os.path.join(
    os.path.dirname(__file__),
    'ujenkins/__init__.py'
)

with open(init_file_path, encoding='utf-8') as f:
    try:
        version = re.findall(r"__version__ = '(.*)'", f.read())[0]
    except IndexError:
        raise RuntimeError('Unable to get package version')

with open('README.rst', encoding='utf-8') as f:
    README = f.read()

setup_args = dict(
    name='ujenkins',
    version=version,
    description='Universal (sync/async) Python client for Jenkins',
    long_description_content_type='text/x-rst',
    long_description=README,
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    license='MIT',
    packages=find_packages(),
    package_data={'ujenkins': ['*']},
    author='Petr Belskiy',
    keywords=['jenkins'],
    url='https://github.com/pbelskiy/ujenkins',
    download_url='https://pypi.org/project/ujenkins'
)

install_requires = [
    'aiohttp>3.6.2,<4.0.0',
    'requests>=2.24.0,<3.0.0',
]

setup(
    install_requires=install_requires,
    python_requires='>=3.5',
    **setup_args
)



from http import HTTPStatus
from typing import Any, Callable, List, Optional, Union

from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from ujenkins.adapters import CRUMB_ISSUER_URL
from ujenkins.core import Jenkins, Response
from ujenkins.exceptions import JenkinsError, JenkinsNotFoundError


class JenkinsClient(Jenkins):

    def __init__(self,
                 url: str,
                 user: Optional[str] = None,
                 password: Optional[str] = None,
                 *,
                 verify: bool = True,
                 timeout: Optional[float] = None,
                 retry: Optional[dict] = None
                 ):
        """
        Jenkins sync client class.

        Args:
            url (str):
                URL of Jenkins server.

            user (Optional[str]):
                User name.

            password (Optional[str]):
                Password for user.

            verify (Optional[bool]):
                Verify SSL (default: true).

            timeout (Optional[float]):
                HTTP request timeout.

            retry (Optional[dict]):
                Retry options to prevent failures if server restarting or
                temporary network problem. Disabled by default use total > 0
                to enable.

                - total: ``int`` Total retries count.
                - factor: ``int`` Sleep between retries (default 1)
                    {factor} * (2 ** ({number of total retries} - 1))
                - statuses: ``List[int]`` HTTP statues retries on. (default [])

                Example:

                .. code-block:: python

                    retry = dict(
                        total=10,
                        factor=1,
                        statuses=[500]
                    )

                With factor = 1

                ============  =============
                Retry number  Sleep
                ============  =============
                1              0.5 seconds
                2              1.0 seconds
                3              2.0 seconds
                4              4.0 seconds
                5              8.0 seconds
                6             16.0 seconds
                7             32.0 seconds
                8              1.1 minutes
                9              2.1 minutes
                10             4.3 minutes
                11             8.5 minutes
                12            17.1 minutes
                13            34.1 minutes
                14             1.1 hours
                15             2.3 hours
                16             4.6 hours
                17             9.1 hours
                18            18.2 hours
                19            36.4 hours
                20            72.8 hours
                ============  =============

        Returns:
            Client instance
        """
        super().__init__()

        self.host = url.rstrip('/')
        self.session = Session()

        if user and password:
            self.session.auth = (user, password)

        self.timeout = timeout
        self.verify = verify
        self.crumb = None  # type: Any

        if not retry:
            return

        self._validate_retry_argument(retry)
        adapter = HTTPAdapter(max_retries=Retry(
            total=retry['total'],
            backoff_factor=retry.get('factor', 1),
            status_forcelist=retry.get('statuses', []),
            allowed_methods=['GET', 'POST', 'PATCH'],
        ))

        self.session.mount('http://', adapter)
        self.session.mount('https://', adapter)

    def _http_request(self,
                      method: str,
                      path: str,
                      *,
                      _callback: Optional[Callable] = None,
                      **kwargs: Any
                      ) -> Any:

        if self.timeout and 'timeout' not in kwargs:
            kwargs['timeout'] = self.timeout

        if self.crumb:
            kwargs.setdefault('headers', {})
            kwargs['headers'].update(self.crumb)

        response = self.session.request(
            method,
            f'{self.host}{path}',
            verify=self.verify,
            **kwargs
        )

        result = self._process(
            Response(response.status_code, response.headers, response.text),
            _callback
        )

        return result

    def _get_crumb(self) -> Union[bool, dict]:
        try:
            response = self._http_request('GET', CRUMB_ISSUER_URL)
            self.crumb = {response['crumbRequestField']: response['crumb']}
            return self.crumb
        except JenkinsNotFoundError:
            return False

    def _request(self,
                 method: str,
                 path: str,
                 **kwargs: Any
                 ) -> Any:
        """
        Core class method for endpoints, which wraps auto crumb detection.
        """
        if self.crumb:
            try:
                return self._http_request(method, path, **kwargs)
            except JenkinsError as e:
                if e.status != HTTPStatus.FORBIDDEN:
                    raise

        if self.crumb is not False:
            self.crumb = self._get_crumb()

        return self._http_request(method, path, **kwargs)

    @staticmethod
    def _chain(functions: List[Callable]) -> Any:
        """
        Helper function for creating call chain for functions.
        """
        prev = None

        for func in functions:
            try:
                prev = func(prev)

                while True:
                    if callable(prev):
                        prev = prev()
                    else:
                        break
            except JenkinsError as e:
                prev = e

        return prev

    def close(self) -> None:
        """
        Close client session
        """
        self.session.close()
